{% extends "base.html" %}

{% block extrahead %}
  <!-- Google Site Verification -->
  <meta name="google-site-verification" content="p_z6B5TiMdq2eo0Nv7iOWaYanUBHwf2pxQ5B-qvbE_o">
  
  <!-- Sitemap Reference -->
  <link rel="sitemap" type="application/xml" title="Sitemap" href="{{ config.site_url }}/sitemap.xml">
  
  <!-- Referrer Policy -->
  <meta name="referrer" content="no-referrer-when-downgrade">
  
  <!-- Home Page Reference -->
  <link rel="home" href="{{ config.site_url }}/" title="AutoTimm Home">
  
  <!-- SEO Meta Tags -->
  <meta name="description" content="AutoTimm - Automatic PyTorch Image Models for Computer Vision. Build, train, and deploy state-of-the-art models for image classification, object detection, and segmentation with minimal code.">
  <meta name="keywords" content="PyTorch, Computer Vision, Deep Learning, Image Classification, Object Detection, Semantic Segmentation, Instance Segmentation, Transfer Learning, timm, YOLOX, AutoML, Machine Learning, Neural Networks, CNN, Vision Transformer, Model Training, Model Deployment, TorchScript">
  <meta name="author" content="Krishnatheja Vanka">
  <meta name="robots" content="index, follow">
  
  <!-- Open Graph Meta Tags -->
  <meta property="og:type" content="website">
  <meta property="og:title" content="AutoTimm - Automatic PyTorch Image Models">
  <meta property="og:description" content="Build, train, and deploy state-of-the-art computer vision models with minimal code. Support for classification, detection, and segmentation tasks.">
  <meta property="og:url" content="https://theja-vanka.github.io/AutoTimm">
  <meta property="og:site_name" content="AutoTimm">
  <meta property="og:image" content="https://theja-vanka.github.io/AutoTimm/assets/logo.png">
  
  <!-- Twitter Card Meta Tags -->
  <meta name="twitter:card" content="summary_large_image">
  <meta name="twitter:title" content="AutoTimm - Automatic PyTorch Image Models">
  <meta name="twitter:description" content="Build, train, and deploy state-of-the-art computer vision models with minimal code.">
  <meta name="twitter:image" content="https://theja-vanka.github.io/AutoTimm/assets/logo.png">
  
  <!-- Structured Data for Breadcrumbs -->
  <script type="application/ld+json">
  {
    "@context": "https://schema.org",
    "@type": "WebSite",
    "name": "{{ config.site_name }}",
    "url": "{{ config.site_url }}",
    "description": "{{ config.site_description }}",
    "publisher": {
      "@type": "Person",
      "name": "{{ config.site_author }}"
    }
  }
  </script>
{% endblock %}
